package main

import (
	"encoding/json"
	"fmt"
	log "github.com/Sirupsen/logrus"
	"github.com/hyperledger/fabric/core/chaincode/shim"
	pb "github.com/hyperledger/fabric/protos/peer"
	"gitlab.smartblocklab.com/smartcontracts/common/security/checker"
	"gitlab.smartblocklab.com/smartcontracts/common/security/actions"
	"gitlab.smartblocklab.com/smartcontracts/common/security/requests"
)

const (
	AddAuthInfo    = "addAuthInfo"
	GetAuthInfo    = "getAuthInfo"
	UpdateAuthInfo = "updateAuthInfo"
)

var (
	logger = log.WithFields(log.Fields{
		"module":    "auth",
		"chaincode": "authCC",
	})
)

type Auth struct {
	UserID   string `json:"user_id"`
	Login    string `json:"login,omitempty"`
	Verifier string `json:"verifier,omitempty"`
	Salt     []byte `json:"salt,omitempty"`
	PrivateSalt string `json:"private_salt,omitempty"`
	PublicSalt string `json:"public_salt,omitempty"`
}

func (authInfo Auth) Marshal() pb.Response {
	bytes, err := json.Marshal(authInfo)
	if err != nil {
		errMsg := fmt.Sprintf("Failed to serialize authInfo. Error: %s", err)
		logger.Error(errMsg)
		return shim.Error(errMsg)
	}

	return shim.Success(bytes)
}

type ChaincodeStubInterfaceAdapter interface {
	GetState(key string) ([]byte, error)

	PutState(key string, value []byte) error

	GetQueryResult(query string) (shim.StateQueryIteratorInterface, error)

	InvokeChaincode(chaincodeName string, args [][]byte, channel string) pb.Response
}

type AuthChaincodeAction func(args []string, stub ChaincodeStubInterfaceAdapter) pb.Response

type AuthChaincode struct {
	actions map[string]AuthChaincodeAction
}

func (t *AuthChaincode) Init(stub shim.ChaincodeStubInterface) pb.Response {
	logger.Info("Start initializing chaincode...")
	return shim.Success([]byte("Chaincode initialized successfully"))

}

//Invoke metods of Blockchain
func (t *AuthChaincode) Invoke(stub shim.ChaincodeStubInterface) pb.Response {
	actionName, args := stub.GetFunctionAndParameters()
	logger.Info("Invoke function " + actionName)

	if action, ok := t.actions[actionName]; ok {
		return action(args, stub)
	}

	// No action defined with provided function name, reporting error
	invalidFuncName := fmt.Sprintf("Invalid invoke function name, function name provided %s ", actionName)
	logger.Error(invalidFuncName)
	return shim.Error(invalidFuncName)
}

func (t *AuthChaincode) addAuthInfo(args []string, stub ChaincodeStubInterfaceAdapter) pb.Response {
	logger.Debug("AuthChaincode::addAuthInfo ENTER")
	defer logger.Debug("AuthChaincode::addAuthInfo EXIT")

	_, systemParams, err := checker.CheckAccess(stub, args, actions.Actions().AuthCC().AddAuthInfo)
	if err != nil {
		errMsg := fmt.Sprintf("SECURITY BREACH: %s", err.Error())
		logger.Error(errMsg)
		return shim.Error(errMsg)
	}

	var auth Auth

	err = json.Unmarshal(systemParams, &auth)
	if err != nil {
		return shim.Error(fmt.Sprintf("Bad arg json: %s ", err.Error()))
	}

	if len(auth.UserID) == 0 {
		errMsg := fmt.Sprintf("Missing user id in Auth record, information received %s", args[0])
		logger.Error(errMsg)
		return shim.Error(errMsg)
	}

	isExists := t.isExists(auth.UserID, stub)
	if isExists {
		errMsg := fmt.Sprintf("Can't add Auth, due to record with same UserID %s already exists", auth.UserID)
		logger.Error(errMsg)
		return shim.Error(errMsg)
	}

	// можно использовать JSON который пришел с запросом, потому что мы
	// ничего не меняем перед вставкой
	err = stub.PutState(auth.UserID, []byte(args[0]))
	if err != nil {
		errMsg := fmt.Sprintf("Could not store auth record information, auth id %s, due to %s", auth.UserID, err)
		logger.Error(errMsg)
		return shim.Error(errMsg)
	}

	return shim.Success([]byte(args[0]))
}

func (t *AuthChaincode) updateAuthInfo(args []string, stub ChaincodeStubInterfaceAdapter) pb.Response {
	logger.Debug("AuthChaincode::updateAuthInfo ENTER")
	defer logger.Debug("AuthChaincode::updateAuthInfo EXIT")

	_, systemParams, err := checker.CheckAccess(stub, args, actions.Actions().AuthCC().UpdateAuthInfo)
	if err != nil {
		errMsg := fmt.Sprintf("SECURITY BREACH: %s", err.Error())
		logger.Error(errMsg)
		return shim.Error(errMsg)
	}

	var auth Auth

	err = json.Unmarshal(systemParams, &auth)
	if err != nil {
		return shim.Error(fmt.Sprintf("Bad arg json: %s ", err.Error()))
	}

	if len(auth.UserID) == 0 {
		errMsg := fmt.Sprintf("Missing user id in Auth record, auth information received %s", args[0])
		logger.Error(errMsg)
		return shim.Error(errMsg)
	}

	isExists := t.isExists(auth.UserID, stub)
	if !isExists {
		errMsg := fmt.Sprintf("Can't update, due to record Auth with UserID %s does not exists", auth.UserID)
		logger.Error(errMsg)
		return shim.Error(errMsg)
	}

	err = stub.PutState(auth.UserID, []byte(args[0]))
	if err != nil {
		errMsg := fmt.Sprintf("Could not store auth record information, auth id %s, due to %s", auth.UserID, err)
		logger.Error(errMsg)
		return shim.Error(errMsg)
	}

	return shim.Success([]byte(args[0]))
}

func (t *AuthChaincode) getAuthInfo(args []string, stub ChaincodeStubInterfaceAdapter) pb.Response {
	logger.Debug("AuthChaincode::getAuthInfo ENTER")
	defer logger.Debug("AuthChaincode::getAuthInfo EXIT")

	_, systemParams, err := checker.CheckAccess(stub, args, actions.Actions().AuthCC().GetAuthInfo)
	if err != nil {
		errMsg := fmt.Sprintf("SECURITY BREACH: %s", err.Error())
		logger.Error(errMsg)
		return shim.Error(errMsg)
	}

	var userID requests.Id

	err = json.Unmarshal(systemParams, &userID)
	if err != nil {
		return shim.Error(fmt.Sprintf("Bad arg json: %s ", err.Error()))
	}

	authRecord, err := stub.GetState(userID.Id)
	if err != nil {
		errMsg := fmt.Sprintf("Error reading previous auth record, due to %s", err)
		logger.Error(errMsg)
		return shim.Error(errMsg)
	}

	return shim.Success(authRecord)
}

func (t *AuthChaincode) isExists(userID string, stub ChaincodeStubInterfaceAdapter) bool {
	authRecord, err := stub.GetState(userID)
	if err != nil {
		errMsg := fmt.Sprintf("Error reading previous auth record, due to %s", err)
		logger.Error(errMsg)
		return false
	}

	var auth Auth
	err = json.Unmarshal(authRecord, &auth)
	if err != nil {
		errMsg := fmt.Sprintf("Wasn't able to read auth information passed as paraemter, due to %s", err)
		logger.Error(errMsg)
		return false
	}

	return true
}

// NewAuthChaincode created new instance of the Auth chaincode
func NewAuthChaincode() *AuthChaincode {
	chaincode := &AuthChaincode{}
	chaincode.actions = map[string]AuthChaincodeAction{
		AddAuthInfo:    chaincode.addAuthInfo,
		GetAuthInfo:    chaincode.getAuthInfo,
		UpdateAuthInfo: chaincode.updateAuthInfo,
	}
	return chaincode
}
