package main

import (
	"github.com/hyperledger/fabric/core/chaincode/shim"
	"github.com/stretchr/testify/mock"
	"github.com/hyperledger/fabric/protos/ledger/queryresult"
	"testing"
	"github.com/stretchr/testify/assert"
	"encoding/json"
)

// ChaincodeStubInterfaceAdapterMock chaincode adapter mock interface
type ChaincodeStubInterfaceAdapterMock struct {
	mock.Mock
}

// GetState
func (mock *ChaincodeStubInterfaceAdapterMock) GetState(key string) ([]byte, error) {
	args := mock.Called(key)
	if args.Get(0) == nil && args.Get(1) == nil {
		return nil, nil
	}
	return args.Get(0).([]byte), args.Error(1)
}

// PutState
func (mock *ChaincodeStubInterfaceAdapterMock) PutState(key string, value []byte) error {
	args := mock.Called(key, value)
	return args.Error(0)
}

// GetQueryResult
func (mock *ChaincodeStubInterfaceAdapterMock) GetQueryResult(query string) (shim.StateQueryIteratorInterface, error) {
	args := mock.Called(query)
	return args.Get(0).(shim.StateQueryIteratorInterface), args.Error(1)
}

// stateQueryIteratorInterfaceMock query result iterator
type stateQueryIteratorInterfaceMock struct {
	mock.Mock
}

func (query *stateQueryIteratorInterfaceMock) Close() error {
	return nil
}

func (query *stateQueryIteratorInterfaceMock) Next() (*queryresult.KV, error) {
	return query.Called().Get(0).(*queryresult.KV), query.Called().Error(1)
}

func (query *stateQueryIteratorInterfaceMock) HasNext() bool {
	return query.Called().Get(0).(bool)
}

func TestNewAuthChaincode_addAuthInfo_VALID(t *testing.T) {
	auth := &Auth{
		UserID:         "testUserID",
		Login:          "testLogin",
		HashedPassword: "testHashedPassword",
		Salt:           []byte("testSalt"),
		Verifier:       "testVerifier",
	}

	a := assert.New(t)

	authJson, err := json.Marshal(auth)
	a.NoError(err)

	stubAdapter := &ChaincodeStubInterfaceAdapterMock{}
	stubAdapter.On("GetState", "testWalletID").Return(nil, nil)
	stubAdapter.On("PutState", "testWalletID", mock.Anything).Run(func(args mock.Arguments) {

	}).Return(nil) // No error

	// Create new account chaincode
	authChaincode := NewAuthChaincode()
	resp := authChaincode.addAuthInfo([]string{string(authJson)}, stubAdapter)

	a.Equal(int32(shim.OK), resp.Status)
}

func TestNewAuthChaincode_updateAuthInfo_VALID(t *testing.T) {
	auth := &Auth{
		UserID:         "testUserID",
		Login:          "testLogin",
		HashedPassword: "testHashedPassword",
		Salt:           []byte("testSalt"),
		Verifier:       "testVerifier",
	}

	a := assert.New(t)

	authJson, err := json.Marshal(auth)
	a.NoError(err)

	stubAdapter := &ChaincodeStubInterfaceAdapterMock{}
	stubAdapter.On("GetState", "testUserID").Return(nil, nil)
	stubAdapter.On("PutState", "testWalletID", mock.Anything).Run(func(args mock.Arguments) {

	}).Return(nil) // No error

	// Create new account chaincode
	authChaincode := NewAuthChaincode()
	resp := authChaincode.updateAuthInfo([]string{string(authJson)}, stubAdapter)

	a.Equal(int32(shim.OK), resp.Status)
}

func TestNewAuthChaincode_getAuthInfo_VALID(t *testing.T) {
	auth := &Auth{
		UserID:         "testUserID",
		Login:          "testLogin",
		HashedPassword: "testHashedPassword",
		Salt:           []byte("testSalt"),
		Verifier:       "testVerifier",
	}

	a := assert.New(t)

	authJson, err := json.Marshal(auth)
	a.NoError(err)

	stubAdapter := &ChaincodeStubInterfaceAdapterMock{}
	stubAdapter.On("GetState", "testUserID").Return(nil, nil)

	// Create new account chaincode
	authChaincode := NewAuthChaincode()
	resp := authChaincode.getAuthInfo([]string{string(authJson)}, stubAdapter)

	a.Equal(int32(shim.OK), resp.Status)
}
